# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 28.4 `univar` Chapter.

``univar`` - Univariate Estimation

This module deals with the estimation of unknown parameters of a
univariate distribution.
It includes both point and interval estimation using maximum likelihood
and robust methods.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_28.4/flhtml/g07/g07intro.html

See Also
--------
``naginterfaces.library.examples.univar`` :
    This subpackage contains examples for the ``univar`` module.
    See also the :ref:`library_univar_ex` subsection.
"""

# NAG Copyright 2017-2022.

import collections as _collections

from ..base import univar as _b_univar
from ..base import utils as _b_utils
from ..library import _utils

CiBinomialReturnData = _collections.namedtuple(
    'CiBinomialReturnData',
    ['pl', 'pu'],
)

def ci_binomial(n, k, clevel):
    u"""
    Computes confidence interval for the parameter of a binomial
    distribution.

    ``ci_binomial`` computes a confidence interval for the parameter p
    (the probability of a success) of a binomial distribution.

    For full information please refer to the NAG Library document for
    g07aa

    https://www.nag.com/numeric/nl/nagdoc_28.4/flhtml/g07/g07aaf.html

    Parameters
    ----------
    n : int
        n, the number of trials.

    k : int
        k, the number of successes.

    clevel : float
        The confidence level, (1-alpha), for two-sided interval
        estimate. For example `clevel` = 0.95 will give a 95% confidence
        interval.

    Returns
    -------
    pl : float
        The lower limit, p_l, of the confidence interval.

    pu : float
        The upper limit, p_u, of the confidence interval.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `clevel` = *<value>*.

            Constraint: 0.0 < `clevel` < 1.0.

        (`errno` 1)
            On entry, `n` = *<value>* and `k` = *<value>*.

            Constraint: `n` >= `k`.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            When using the relationship with the gamma distribution the
            series to calculate the gamma probabilities has failed to
            converge.

    Notes
    -----
    Given the number of trials, n, and the number of successes, k, this
    function computes a 100*(1-alpha)% confidence interval for p, the
    probability parameter of a binomial distribution with probability
    function,

        f(x) = (n;x)p^x(1-p)^(n-x), x = 0,1,...,n,

    where alpha is in the interval (0,1).

    Let the confidence interval be denoted by [p_l,p_u].

    The point estimate for p is p^= k/n.

    The lower and upper confidence limits p_l and p_u are estimated by
    the solutions to the equations;

        sum_x = k^n(n;x)p_l^x(1-p_l)^(n-x) = alpha/2,

        sum_x = 0^k(n;x)p_u^x(1-p_u)^(n-x) = alpha/2.

    Three different methods are used depending on the number of trials,
    n, and the number of successes, k.

    (1) If max(k, n-k) < 10^6.

        The relationship between the beta and binomial distributions
        (see page 38 of Hastings and Peacock (1975)) is used to derive
        the equivalent equations,

            [table omitted]

        where beta_{a,b,delta} is the deviate associated with the lower
        tail probability, delta, of the beta distribution with
        parameters a and b.
        These beta deviates are computed using ``stat.inv_cdf_beta``.

    (#) If max(k, n-k) >= 10^6 and min(k, n-k) <= 1000.

        The binomial variate with parameters n and p is approximated by
        a Poisson variate with mean np, see page 38 of Hastings and
        Peacock (1975).

        The relationship between the Poisson and chi^2-distributions
        (see page 112 of Hastings and Peacock (1975)) is used to derive
        the following equations;

            [table omitted]

        where chi_delta,nu^2 is the deviate associated with the lower
        tail probability, delta, of the chi^2-distribution with nu
        degrees of freedom.

        In turn the relationship between the chi^2-distribution and the
        gamma distribution (see page 70 of Hastings and Peacock (1975))
        yields the following equivalent equations;

            [table omitted]

        where gamma_{alpha,beta;delta} is the deviate associated with
        the lower tail probability, delta, of the gamma distribution
        with shape parameter alpha and scale parameter beta.
        These deviates are computed using ``stat.inv_cdf_gamma``.

    (#) If max(k, n-k) > 10^6 and min(k, n-k) > 1000.

        The binomial variate with parameters n and p is approximated by
        a Normal variate with mean np and variance np(1-p), see page 38
        of Hastings and Peacock (1975).

        The approximate lower and upper confidence limits p_l and p_u
        are the solutions to the equations;

            [table omitted]

        where z_delta is the deviate associated with the lower tail
        probability, delta, of the standard Normal distribution.
        These equations are solved using a quadratic equation solver
        (``zeros.quadratic_real``).

    References
    ----------
    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth

    Snedecor, G W and Cochran, W G, 1967, `Statistical Methods`, Iowa
    State University Press
    """

    _return_scalars = _b_univar.ci_binomial(
        n,
        k,
        clevel,
    )
    _pl_py = _return_scalars.pl
    _pu_py = _return_scalars.pu
    _return_data = CiBinomialReturnData(
        _pl_py,
        _pu_py,
    )
    return _return_data

CiPoissonReturnData = _collections.namedtuple(
    'CiPoissonReturnData',
    ['tl', 'tu'],
)

def ci_poisson(n, xmean, clevel):
    u"""
    Computes confidence interval for the parameter of a Poisson
    distribution.

    ``ci_poisson`` computes a confidence interval for the mean parameter
    of the Poisson distribution.

    For full information please refer to the NAG Library document for
    g07ab

    https://www.nag.com/numeric/nl/nagdoc_28.4/flhtml/g07/g07abf.html

    Parameters
    ----------
    n : int
        n, the sample size.

    xmean : float
        The sample mean, x_.

    clevel : float
        The confidence level, (1-alpha), for two-sided interval
        estimate. For example `clevel` = 0.95 gives a 95% confidence
        interval.

    Returns
    -------
    tl : float
        The lower limit, theta_l, of the confidence interval.

    tu : float
        The upper limit, theta_u, of the confidence interval.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `clevel` = *<value>*.

            Constraint: 0.0 < `clevel` < 1.0.

        (`errno` 1)
            On entry, `xmean` = *<value>*.

            Constraint: `xmean` >= 0.0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 2)
            When using the relationship with the gamma distribution the
            series to calculate the gamma probabilities has failed to
            converge.

    Notes
    -----
    Given a random sample of size n, denoted by x_1,x_2,...,x_n, from a
    Poisson distribution with probability function

        p(x) = e^(-theta)(theta^x)/x!, x = 0,1,2,...

    the point estimate, theta^, for theta is the sample mean, x_.

    Given n and x_ this function computes a 100*(1-alpha)% confidence
    interval for the parameter theta, denoted by [theta_l,theta_u],
    where alpha is in the interval (0,1).

    The lower and upper confidence limits are estimated by the solutions
    to the equations

        e^(-n theta_l)sum_x = T^infinity ((n theta_l)^x)/x! =
        alpha/2,;;;e^(-n theta_u)sum_x = 0^T((n theta_u)^x)/x! =
        alpha/2,

    where T = sum_i = 1^nx_i = n theta^.

    The relationship between the Poisson distribution and the
    chi^2-distribution (see page 112 of Hastings and Peacock (1975)) is
    used to derive the equations

        theta_l = 1/2n chi_2T,alpha/2^2,;;theta_u = 1/2n
        chi_2T+2,1-alpha/2^2,

    where chi_nu,p^2 is the deviate associated with the lower tail
    probability p of the chi^2-distribution with nu degrees of freedom.

    In turn the relationship between the chi^2-distribution and the
    gamma distribution (see page 70 of Hastings and Peacock (1975))
    yields the following equivalent equations;

        theta_l = 1/2n gamma_{T,2;alpha/2},;;theta_u = 1/2n
        gamma_{T+1,2;1-alpha/2},

    where gamma_{alpha,beta;delta} is the deviate associated with the
    lower tail probability, delta, of the gamma distribution with shape
    parameter alpha and scale parameter beta.
    These deviates are computed using ``stat.inv_cdf_gamma``.

    References
    ----------
    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth

    Snedecor, G W and Cochran, W G, 1967, `Statistical Methods`, Iowa
    State University Press
    """

    _return_scalars = _b_univar.ci_poisson(
        n,
        xmean,
        clevel,
    )
    _tl_py = _return_scalars.tl
    _tu_py = _return_scalars.tu
    _return_data = CiPoissonReturnData(
        _tl_py,
        _tu_py,
    )
    return _return_data

EstimNormalReturnData = _collections.namedtuple(
    'EstimNormalReturnData',
    ['xmu', 'xsig', 'sexmu', 'sexsig', 'corr', 'dev', 'nobs', 'nit'],
)

def estim_normal(method, x, xc, ic, xmu, xsig, tol, maxit):
    u"""
    Computes maximum likelihood estimates for parameters of the Normal
    distribution from grouped and/or censored data.

    ``estim_normal`` computes maximum likelihood estimates and their
    standard errors for parameters of the Normal distribution from
    grouped and/or censored data.

    For full information please refer to the NAG Library document for
    g07bb

    https://www.nag.com/numeric/nl/nagdoc_28.4/flhtml/g07/g07bbf.html

    Parameters
    ----------
    method : str, length 1
        Indicates whether the Newton--Raphson or EM algorithm should be
        used.

        If `method` = 'N', the Newton--Raphson algorithm is used.

        If `method` = 'E', the EM algorithm is used.

    x : float, array-like, shape (n)
        The observations x_i, L_i or U_i, for i = 1,2,...,n.

        If the observation is exactly specified -- the exact value, x_i.

        If the observation is right-censored -- the lower value, L_i.

        If the observation is left-censored -- the upper value, U_i.

        If the observation is interval-censored -- the lower or upper
        value, L_i or U_i, (see `xc`).

    xc : float, array-like, shape (n)
        If the jth observation, for j = 1,2,...,n is an
        interval-censored observation then `xc`[j-1] should contain the
        complementary value to `x`[j-1], that is, if `x`[j-1] <
        `xc`[j-1], then `xc`[j-1] contains upper value, U_i, and if
        `x`[j-1] > `xc`[j-1], then `xc`[j-1] contains lower value, L_i.
        Otherwise if the jth observation is exact or right - or
        left-censored `xc`[j-1] need not be set.

        Note: if `x`[j-1] = `xc`[j-1] then the observation is ignored.

    ic : int, array-like, shape (n)
        `ic`[i-1] contains the censoring codes for the ith observation,
        for i = 1,2,...,n.

        If `ic`[i-1] = 0, the observation is exactly specified.

        If `ic`[i-1] = 1, the observation is right-censored.

        If `ic`[i-1] = 2, the observation is left-censored.

        If `ic`[i-1] = 3, the observation is interval-censored.

    xmu : float
        If `xsig` > 0.0 the initial estimate of the mean, mu; otherwise
        `xmu` need not be set.

    xsig : float
        Specifies whether an initial estimate of mu and sigma are to be
        supplied.

        `xsig` > 0.0

            `xsig` is the initial estimate of sigma and `xmu` must
            contain an initial estimate of mu.

        `xsig` <= 0.0

            Initial estimates of `xmu` and `xsig` are calculated
            internally from:

            (a) the exact observations, if the number of exactly
                specified observations is >= 2; or

            (#) the interval-censored observations; if the number of
                interval-censored observations is >= 1; or

            (#) they are set to 0.0 and 1.0 respectively.

    tol : float
        The relative precision required for the final estimates of mu
        and sigma. Convergence is assumed when the absolute relative
        changes in the estimates of both mu and sigma are less than
        `tol`.

        If `tol` = 0.0, a relative precision of 0.000005 is used.

    maxit : int
        The maximum number of iterations.

        If `maxit` <= 0, a value of 25 is used.

    Returns
    -------
    xmu : float
        The maximum likelihood estimate, mu^, of mu.

    xsig : float
        The maximum likelihood estimate, sigma^, of sigma.

    sexmu : float
        The estimate of the standard error of mu^.

    sexsig : float
        The estimate of the standard error of sigma^.

    corr : float
        The estimate of the correlation between mu^ and sigma^.

    dev : float
        The maximized log-likelihood, L(mu^,sigma^).

    nobs : int, ndarray, shape (4)
        The number of the different types of each observation;

        `nobs`[0] contains number of right-censored observations.

        `nobs`[1] contains number of left-censored observations.

        `nobs`[2] contains number of interval-censored observations.

        `nobs`[3] contains number of exactly specified observations.

    nit : int
        The number of iterations performed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'N' or 'E'.

        (`errno` 1)
            On entry, effective number of observations < 2.

        (`errno` 1)
            On entry, i = *<value>* and `ic`[i-1] = *<value>*.

            Constraint: `ic`[i-1] = 0, 1, 2 or 3.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: machine precision < `tol` <= 1.0 or `tol` = 0.0.

        (`errno` 2)
            The chosen method has not converged in *<value>* iterations.

        (`errno` 3)
            The process has diverged.

        (`errno` 3)
            The EM process has failed.

        (`errno` 4)
            Standard errors cannot be computed.

    Notes
    -----
    A sample of size n is taken from a Normal distribution with mean mu
    and variance sigma^2 and consists of grouped and/or censored data.
    Each of the n observations is known by a pair of values (L_i,U_i)
    such that:

        L_i <= x_i <= U_i.

    The data is represented as particular cases of this form:

        exactly specified observations occur when L_i = U_i = x_i,

        right-censored observations, known only by a lower bound, occur
        when U_i→ infinity,

        left-censored observations, known only by a upper bound, occur
        when L_i→ - infinity,

        and interval-censored observations when L_i < x_i < U_i.

    Let the set A identify the exactly specified observations, sets B
    and C identify the observations censored on the right and left
    respectively, and set D identify the observations confined between
    two finite limits.
    Also let there be r exactly specified observations, i.e., the number
    in A.
    The probability density function for the standard Normal
    distribution is

        Z(x) = 1/(sqrt(2 pi))exp(-1/2x^2), - infinity < x < infinity

    and the cumulative distribution function is

        P(X) = 1-Q(X) = int_- infinity^XZ(x)dx.

    The log-likelihood of the sample can be written as:

        L(mu,sigma) = -rlog(sigma)-1/2sum_A{(x_i-mu)/sigma}^2+sum_B
        log(Q(l_i))+sum_C log(P(u_i))+sum_D log(p_i)

    where p_i = P(u_i)-P(l_i) and u_i = (U_i-mu)/sigma, l_i =
    (L_i-mu)/sigma.

    Let

        S(x_i) = Z(x_i)/Q(x_i), S_1(l_i,u_i) = (Z(l_i)-Z(u_i))/(p_i)

    and

        S_2(l_i,u_i) = (u_i Z(u_i)-l_i Z(l_i))/(p_i),

    then the first derivatives of the log-likelihood can be written as:

        dL(mu,sigma)/d mu = L_1(mu,sigma) =
        sigma^-2sum_A(x_i-mu)+sigma^-1sum_B S(l_i)-sigma^-1sum_C
        S(-u_i)+sigma^-1sum_D S_1(l_i,u_i)

    and

        dL(mu,sigma)/d sigma = L_2(mu,sigma) = -r
        sigma^-1+sigma^-3sum_A(x_i-mu)^2+sigma^-1sum_B l_i
        S(l_i)-sigma^-1sum_C u_i S(-u_i)

        -sigma^-1sum_D S_2(l_i,u_i)

    The maximum likelihood estimates, mu^ and sigma^, are the solution
    to the equations:

        L_1(mu^,sigma^) = 0

    and

        L_2(mu^,sigma^) = 0

    and if the second derivatives d^2L/d^2 mu, d^2L/d mu dsigma and
    d^2L/d^2 sigma are denoted by L_11, L_12 and L_22 respectively, then
    estimates of the standard errors of mu^ and sigma^ are given by:

        se(mu^) = sqrt((-L_22)/(L_11 L_22-L_12^2)), se(sigma^) =
        sqrt((-L_11)/(L_11 L_22-L_12^2))

    and an estimate of the correlation of mu^ and sigma^ is given by:

        (L_12)/sqrt(L_12 L_22).

    To obtain the maximum likelihood estimates the equations (1) and (2)
    can be solved using either the Newton--Raphson method or the
    Expectation-maximization (EM) algorithm of Dempster `et al.` (1977).

    **Newton--Raphson Method**

    This consists of using approximate estimates mu~ and sigma~ to
    obtain improved estimates mu~+deltamu~ and sigma~+deltasigma~ by
    solving

        deltamu~L_11+deltasigma~L_12+L_1 =
        0,;;deltamu~L_12+deltasigma~L_22+L_2 = 0,

    for the corrections deltamu~ and deltasigma~.

    **EM Algorithm**

    The expectation step consists of constructing the variable w_i as
    follows:

        if i in A, w_i = x_i

        if i in B, w_i = E(x_i | x_i > L_i) = mu+sigma S(l_i)

        if i in C, w_i = E(x_i | x_i < U_i) = mu-sigma S(-u_i)

        if i in D, w_i = E(x_i | L_i < x_i < U_i) = mu+sigma
        S_1(l_i,u_i)

    the maximization step consists of substituting (3), (4), (5) and (6)
    into (1) and (2) giving:

        mu^= sum_i = 1^nw^_i/n

    and

        sigma^^2 = sum_i = 1^n(w^_i-mu^)^2/{r+sum_B T(l^_i)+sum_C
        T(-u^_i)+sum_D T_1(l^_i,u^_i)}

    where

        T(x) = S(x){S(x)-x}, T_1(l,u) = S_1^2(l,u)+S_2(l,u)

    and where w^_i, l^_i and u^_i are w_i, l_i and u_i evaluated at mu^
    and sigma^.
    Equations (3) and (8) are the basis of the EM iterative procedure
    for finding mu^ and sigma^^2.
    The procedure consists of alternately estimating mu^ and sigma^^2
    using (7) and (8) and estimating {w^_i} using (3) and (6).

    In choosing between the two methods a general rule is that the
    Newton--Raphson method converges more quickly but requires good
    initial estimates whereas the EM algorithm converges slowly but is
    robust to the initial values.
    In the case of the censored Normal distribution, if only a small
    proportion of the observations are censored then estimates based on
    the exact observations should give good enough initial estimates for
    the Newton--Raphson method to be used.
    If there are a high proportion of censored observations then the EM
    algorithm should be used and if high accuracy is required the
    subsequent use of the Newton--Raphson method to refine the estimates
    obtained from the EM algorithm should be considered.

    References
    ----------
    Dempster, A P, Laird, N M and Rubin, D B, 1977, `Maximum likelihood
    from incomplete data via the EM algorithm (with discussion)`, J.
    Roy. Statist. Soc. Ser. B (39), 1--38

    Swan, A V, 1969, `Algorithm AS 16. Maximum likelihood estimation
    from grouped and censored normal data`, Appl. Statist. (18),
    110--114

    Wolynetz, M S, 1979, `Maximum likelihood estimation from confined
    and censored normal data`, Appl. Statist. (28), 185--195
    """

    fun_name = 'naginterfaces.library.univar.estim_normal'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _xc = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xc'},
        xc,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ic = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ic'},
        ic,
        copy=False,
    )
    _nobs = _b_utils._EngineIntArrayType.empty_ndarray(
        (4,),
    )
    _return_scalars = _b_univar.estim_normal(
        method,
        _n,
        _x,
        _xc,
        _ic,
        xmu,
        xsig,
        tol,
        maxit,
        _nobs,
    )
    _xmu_py = _return_scalars.xmu
    _xsig_py = _return_scalars.xsig
    _sexmu_py = _return_scalars.sexmu
    _sexsig_py = _return_scalars.sexsig
    _corr_py = _return_scalars.corr
    _dev_py = _return_scalars.dev
    _nobs_py = _nobs
    _nit_py = _return_scalars.nit
    _return_data = EstimNormalReturnData(
        _xmu_py,
        _xsig_py,
        _sexmu_py,
        _sexsig_py,
        _corr_py,
        _dev_py,
        _nobs_py,
        _nit_py,
    )
    return _return_data

EstimWeibullReturnData = _collections.namedtuple(
    'EstimWeibullReturnData',
    ['beta', 'gamma', 'sebeta', 'segam', 'corr', 'dev', 'nit'],
)

def estim_weibull(cens, x, ic, gamma, tol, maxit):
    u"""
    Computes maximum likelihood estimates for parameters of the Weibull
    distribution.

    ``estim_weibull`` computes maximum likelihood estimates for
    parameters of the Weibull distribution from data which may be
    right-censored.

    For full information please refer to the NAG Library document for
    g07be

    https://www.nag.com/numeric/nl/nagdoc_28.4/flhtml/g07/g07bef.html

    Parameters
    ----------
    cens : str, length 1
        Indicates whether the data is censored or non-censored.

        `cens` = 'N'

            Each observation is assumed to be exactly specified. `ic` is
            not referenced.

        `cens` = 'C'

            Each observation is censored according to the value
            contained in `ic`[i-1], for i = 1,2,...,n.

    x : float, array-like, shape (n)
        `x`[i-1] contains the ith observation, x_i, for i = 1,2,...,n.

    ic : int, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `cens`='C': n; otherwise: 1.

        If `cens` = 'C', `ic`[i-1] contains the censoring codes for the
        ith observation, for i = 1,2,...,n.

        If `ic`[i-1] = 0, the ith observation is exactly specified.

        If `ic`[i-1] = 1, the ith observation is right-censored.

        If `cens` = 'N', `ic` is not referenced.

    gamma : float
        Indicates whether an initial estimate of gamma is provided.

        If `gamma` > 0.0, it is taken as the initial estimate of gamma
        and an initial estimate of beta is calculated from this value of
        gamma.

        If `gamma` <= 0.0, initial estimates of gamma and beta are
        calculated, internally, providing the data contains at least two
        distinct exact observations. (If there are only two distinct
        exact observations, the largest observation must not be exactly
        specified.) See [Further Comments] for further details.

    tol : float
        The relative precision required for the final estimates of beta
        and gamma. Convergence is assumed when the absolute relative
        changes in the estimates of both beta and gamma are less than
        `tol`.

        If `tol` = 0.0, a relative precision of 0.000005 is used.

    maxit : int
        The maximum number of iterations allowed.

        If `maxit` <= 0, a value of 25 is used.

    Returns
    -------
    beta : float
        The maximum likelihood estimate, beta^, of beta.

    gamma : float
        Contains the maximum likelihood estimate, gamma^, of gamma.

    sebeta : float
        An estimate of the standard error of beta^.

    segam : float
        An estimate of the standard error of gamma^.

    corr : float
        An estimate of the correlation between beta^ and gamma^.

    dev : float
        The maximized kernel log-likelihood, L(beta^,gamma^).

    nit : int
        The number of iterations performed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: machine precision < `tol` <= 1.0 or `tol` = 0.0.

        (`errno` 1)
            On entry, `cens` = *<value>*.

            Constraint: `cens` = 'N' or 'C'.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            On entry, i = *<value>* and `ic`[i-1] = *<value>*.

            Constraint: `ic`[i-1] = 0 or 1.

        (`errno` 2)
            On entry, i = *<value>* and `x`[i-1] = *<value>*.

            Constraint: `x`[i-1] > 0.0.

        (`errno` 3)
            Unable to calculate initial values.

        (`errno` 3)
            On entry, there are no exactly specified observations.

        (`errno` 4)
            The chosen method has not converged in *<value>* iterations.

        (`errno` 5)
            The process has diverged.

        (`errno` 5)
            Hessian matrix of the Newton--Raphson process is singular.

        (`errno` 6)
            Potential overflow detected.

    Notes
    -----
    ``estim_weibull`` computes maximum likelihood estimates of the
    parameters of the Weibull distribution from exact or right-censored
    data.

    For n realizations, y_i, from a Weibull distribution a value x_i is
    observed such that

        x_i <= y_i.

    There are two situations:

    (a) exactly specified observations, when x_i = y_i

    (#) right-censored observations, known by a lower bound, when x_i <
        y_i.

    The probability density function of the Weibull distribution, and
    hence the contribution of an exactly specified observation to the
    likelihood, is given by:

        f(x;lambda,gamma) = lambdagamma x^(gamma-1)exp(-lambda x^gamma),
        x > 0, for lambda,gamma > 0;

    while the survival function of the Weibull distribution, and hence
    the contribution of a right-censored observation to the likelihood,
    is given by:

        S(x;lambda,gamma) = exp(-lambda x^gamma), x > 0, for
        lambda,gamma > 0.

    If d of the n observations are exactly specified and indicated by i
    in D and the remaining (n-d) are right-censored, then the likelihood
    function, Like (lambda,gamma) is given by

        Like(lambda,gamma) prop (lambdagamma)^d(prod_{i in
        D}x_i^gamma-1)exp(-lambda sum_i = 1^nx_i^gamma).

    To avoid possible numerical instability a different parameterisation
    beta,gamma is used, with beta = log(lambda).
    The kernel log-likelihood function, L(beta,gamma), is then:

        L(beta,gamma) = dlog(gamma)+d beta+(gamma-1)sum_{i in
        D}log(x_i)-e^beta sum_i = 1^nx_i^gamma.

    If the derivatives dL/d beta, dL/d gamma, d^2L/d beta^2, d^2L/d beta
    dgamma and d^2L/d gamma^2 are denoted by L_1, L_2, L_11, L_12 and
    L_22, respectively, then the maximum likelihood estimates, beta^ and
    gamma^, are the solution to the equations:

        L_1(beta^,gamma^) = 0

    and

        L_2(beta^,gamma^) = 0

    Estimates of the asymptotic standard errors of beta^ and gamma^ are
    given by:

        se(beta^) = sqrt((-L_22)/(L_11 L_22-L_12^2)), se(gamma^) =
        sqrt((-L_11)/(L_11 L_22-L_12^2)).

    An estimate of the correlation coefficient of beta^ and gamma^ is
    given by:

        (L_12)/sqrt(L_12 L_22).

    **Note:** if an estimate of the original parameter lambda is
    required, then

        lambda^= exp(beta^) and se(lambda^) = lambda^se(beta^).

    The equations (1) and (2) are solved by the Newton--Raphson
    iterative method with adjustments made to ensure that gamma^> 0.0.

    References
    ----------
    Gross, A J and Clark, V A, 1975, `Survival Distributions:
    Reliability Applications in the Biomedical Sciences`, Wiley

    Kalbfleisch, J D and Prentice, R L, 1980, `The Statistical Analysis
    of Failure Time Data`, Wiley

    See Also
    --------
    :meth:`naginterfaces.library.examples.univar.estim_weibull_ex.main`
    """

    fun_name = 'naginterfaces.library.univar.estim_weibull'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ic = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ic'},
        ic,
        copy=False,
    )
    _return_scalars = _b_univar.estim_weibull(
        cens,
        _n,
        _x,
        _ic,
        gamma,
        tol,
        maxit,
    )
    _gamma_py = _return_scalars.gamma
    _beta_py = _return_scalars.beta
    _sebeta_py = _return_scalars.sebeta
    _segam_py = _return_scalars.segam
    _corr_py = _return_scalars.corr
    _dev_py = _return_scalars.dev
    _nit_py = _return_scalars.nit
    _return_data = EstimWeibullReturnData(
        _beta_py,
        _gamma_py,
        _sebeta_py,
        _segam_py,
        _corr_py,
        _dev_py,
        _nit_py,
    )
    return _return_data

EstimGenparetoReturnData = _collections.namedtuple(
    'EstimGenparetoReturnData',
    ['xi', 'beta', 'asvc', 'obsvc', 'll'],
)

def estim_genpareto(y, optopt):
    u"""
    Estimates parameter values of the generalized Pareto distribution.

    ``estim_genpareto`` estimates parameter values for the generalized
    Pareto distribution by using either moments or maximum likelihood.

    For full information please refer to the NAG Library document for
    g07bf

    https://www.nag.com/numeric/nl/nagdoc_28.4/flhtml/g07/g07bff.html

    Parameters
    ----------
    y : float, array-like, shape (n)
        The n observations y_i, for i = 1,2,...,n, assumed to follow a
        generalized Pareto distribution.

    optopt : int
        Determines the method of estimation, set:

        `optopt` = -2

            For the method of probability-weighted moments.

        `optopt` = -1

            For the method of moments.

        `optopt` = 1

            For maximum likelihood with starting values given by the
            method of moments estimates.

        `optopt` = 2

            For maximum likelihood with starting values given by the
            method of probability-weighted moments.

    Returns
    -------
    xi : float
        The parameter estimate xi^.

    beta : float
        The parameter estimate beta^.

    asvc : float, ndarray, shape (4)
        The variance-covariance of the asymptotic Normal distribution of
        xi^ and beta^. `asvc`[0] contains the variance of xi^; `asvc`[3]
        contains the variance of beta^; `asvc`[1] and `asvc`[2] contain
        the covariance of xi^ and beta^.

    obsvc : float, ndarray, shape (4)
        If maximum likelihood estimates are requested, the observed
        variance-covariance of xi^ and beta^. `obsvc`[0] contains the
        variance of xi^; `obsvc`[3] contains the variance of beta^;
        `obsvc`[1] and `obsvc`[2] contain the covariance of xi^ and
        beta^.

    ll : float
        If maximum likelihood estimates are requested, `ll` contains the
        log-likelihood value L at the end of the optimization; otherwise
        `ll` is set to -1.0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 1.

        (`errno` 2)
            On entry, `y`[*<value>*] = *<value>*.

            Constraint: `y`[i-1] >= 0.0 for all i.

        (`errno` 3)
            On entry, `optopt` = *<value>*.

            Constraint: `optopt` = -2, -1, 1 or 2.

        (`errno` 9)
            The optimization of log-likelihood failed to converge; no
            maximum likelihood estimates are returned. Try using the
            other maximum likelihood option by resetting `optopt`. If
            this also fails, moments-based estimates can be returned by
            an appropriate setting of `optopt`.

        (`errno` 10)
            Variance of data in `y` is too low for method of moments
            optimization.

        (`errno` 11)
            The sum of `y` is zero within machine precision.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            The asymptotic distribution is not available for the
            returned parameter estimates.

        (`errno` 7)
            The distribution of maximum likelihood estimates cannot be
            calculated for the returned parameter estimates because the
            Hessian matrix could not be inverted.

        (`errno` 8)
            The asymptotic distribution of parameter estimates is
            invalid and the distribution of maximum likelihood estimates
            cannot be calculated for the returned parameter estimates
            because the Hessian matrix could not be inverted.

    Notes
    -----
    Let the distribution function of a set of n observations

        y_i, i = 1,2,...,n

    be given by the generalized Pareto distribution:

        F(y) = {[table omitted]

    where

        beta > 0 and

        y >= 0, when xi >= 0;

        0 <= y <= -beta/xi, when xi < 0.

    Estimates xi^ and beta^ of the parameters xi and beta are calculated
    by using one of:

        method of moments (MOM);

        probability-weighted moments (PWM);

        maximum likelihood estimates (MLE) that seek to maximize the
        log-likelihood:

                L = -nln(beta^-(1+1/(xi^))) sum(ln(1+xi^y_i/(beta^))).

    The variances and covariance of the asymptotic Normal distribution
    of parameter estimates xi^ and beta^ are returned if xi^ satisfies:

        xi^< 1/4 for the MOM;

        xi^< 1/2 for the PWM method;

        xi^< -1/2 for the MLE method.

    If the MLE option is exercised, the observed variances and
    covariance of xi^ and beta^ is returned, given by the negative
    inverse Hessian of L.

    References
    ----------
    Hosking, J R M and Wallis, J R, 1987, `Parameter and quantile
    estimation for the generalized Pareto distribution`, Technometrics
    (29(3))

    McNeil, A J, Frey, R and Embrechts, P, 2005, `Quantitative Risk
    Management`, Princeton University Press
    """

    fun_name = 'naginterfaces.library.univar.estim_genpareto'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _asvc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    _obsvc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    try:
        _return_scalars = _b_univar.estim_genpareto(
            _n,
            _y,
            optopt,
            _asvc,
            _obsvc,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _xi_py = _return_scalars.xi
    _beta_py = _return_scalars.beta
    _asvc_py = _asvc
    _obsvc_py = _obsvc
    _ll_py = _return_scalars.ll
    _return_data = EstimGenparetoReturnData(
        _xi_py,
        _beta_py,
        _asvc_py,
        _obsvc_py,
        _ll_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

Ttest2normalReturnData = _collections.namedtuple(
    'Ttest2normalReturnData',
    ['t', 'df', 'prob', 'dl', 'du'],
)

def ttest_2normal(tail, equal, nx, ny, xmean, ymean, xstd, ystd, clevel):
    u"""
    Computes t-test statistic for a difference in means between two
    Normal populations, confidence interval.

    ``ttest_2normal`` computes a t-test statistic to test for a
    difference in means between two Normal populations, together with a
    confidence interval for the difference between the means.

    For full information please refer to the NAG Library document for
    g07ca

    https://www.nag.com/numeric/nl/nagdoc_28.4/flhtml/g07/g07caf.html

    Parameters
    ----------
    tail : str, length 1
        Indicates which tail probability is to be calculated, and thus
        which alternative hypothesis is to be used.

        `tail` = 'T'

            The two tail probability, i.e., H_1:mu_x != mu_y.

        `tail` = 'U'

            The upper tail probability, i.e., H_1:mu_x > mu_y.

        `tail` = 'L'

            The lower tail probability, i.e., H_1:mu_x < mu_y.

    equal : str, length 1
        Indicates whether the population variances are assumed to be
        equal or not.

        `equal` = 'E'

            The population variances are assumed to be equal, that is
            sigma_x^2 = sigma_y^2.

        `equal` = 'U'

            The population variances are not assumed to be equal.

    nx : int
        n_x, the size of the X sample.

    ny : int
        n_y, the size of the Y sample.

    xmean : float
        x_, the mean of the X sample.

    ymean : float
        y_, the mean of the Y sample.

    xstd : float
        s_x, the standard deviation of the X sample.

    ystd : float
        s_y, the standard deviation of the Y sample.

    clevel : float
        The confidence level, 1-alpha, for the specified tail. For
        example `clevel` = 0.95 will give a 95% confidence interval.

    Returns
    -------
    t : float
        Contains the test statistic, t_obs or t'_obs.

    df : float
        Contains the degrees of freedom for the test statistic.

    prob : float
        Contains the significance level, that is the tail probability,
        p, as defined by `tail`.

    dl : float
        Contains the lower confidence limit for mu_x-mu_y.

    du : float
        Contains the upper confidence limit for mu_x-mu_y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `clevel` = *<value>*.

            Constraint: 0.0 < `clevel` < 1.0.

        (`errno` 1)
            On entry, `ystd` = *<value>*.

            Constraint: `ystd` > 0.0.

        (`errno` 1)
            On entry, `xstd` = *<value>*.

            Constraint: `xstd` > 0.0.

        (`errno` 1)
            On entry, `ny` = *<value>*.

            Constraint: `ny` >= 2.

        (`errno` 1)
            On entry, `nx` = *<value>*.

            Constraint: `nx` >= 2.

        (`errno` 1)
            On entry, `equal` = *<value>*.

            Constraint: `equal` = 'E' or 'U'.

        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'T', 'U' or 'L'.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    Consider two independent samples, denoted by X and Y, of size n_x
    and n_y drawn from two Normal populations with means mu_x and mu_y,
    and variances sigma_x^2 and sigma_y^2 respectively.
    Denote the sample means by x_ and y_ and the sample variances by
    s_x^2 and s_y^2 respectively.

    ``ttest_2normal`` calculates a test statistic and its significance
    level to test the null hypothesis H_0:mu_x = mu_y, together with
    upper and lower confidence limits for mu_x-mu_y.
    The test used depends on whether or not the two population variances
    are assumed to be equal.

    (1) It is assumed that the two variances are equal, that is
        sigma_x^2 = sigma_y^2.

        The test used is the two sample t-test.
        The test statistic t is defined by;

            t_obs = (x_-y_)/(s*sqrt((1/n_x)+(1/n_y)))

        where

            s^2 = ((n_x-1)s_x^2+(n_y-1)s_y^2)/(n_x+n_y-2)

        is the pooled variance of the two samples.

        Under the null hypothesis H_0 this test statistic has a
        t-distribution with (n_x+n_y-2) degrees of freedom.

        The test of H_0 is carried out against one of three possible
        alternatives;

        H_1:mu_x != mu_y; the significance level, p = P(t >= |t_obs|),
        i.e., a two tailed probability.

        H_1:mu_x > mu_y; the significance level, p = P(t >= t_obs),
        i.e., an upper tail probability.

        H_1:mu_x < mu_y; the significance level, p = P(t <= t_obs),
        i.e., a lower tail probability.

        Upper and lower 100*(1-alpha)% confidence limits for mu_x-mu_y
        are calculated as:

            (x_-y_)+/-t_{1-alpha/2}s*sqrt((1/n_x)+(1/n_y)).

        where t_{1-alpha/2} is the 100*(1-alpha/2) percentage point of
        the t-distribution with (n_x+n_y-2) degrees of freedom.

    (#) It is not assumed that the two variances are equal.

        If the population variances are not equal the usual two sample
        t-statistic no longer has a t-distribution and an approximate
        test is used.

        This problem is often referred to as the Behrens--Fisher
        problem, see Kendall and Stuart (1969).
        The test used here is based on Satterthwaites procedure.
        To test the null hypothesis the test statistic t' is used where

            t'_obs = (x_-y_)/se(x_-y_)

        where se(x_-y_) = sqrt(s_x^2/(n_x)+s_y^2/(n_y)).

        A t-distribution with f degrees of freedom is used to
        approximate the distribution of t' where

            f =
            se((x_-y_)^4)/(((s_x^2/n_x)^2)/((n_x-1))+((s_y^2/n_y)^2)/((n_y-1))).

        The test of H_0 is carried out against one of the three
        alternative hypotheses described above, replacing t by t' and
        t_obs by t'_obs.

        Upper and lower 100*(1-alpha)% confidence limits for mu_x-mu_y
        are calculated as:

            (x_-y_)+/-t_{1-alpha/2}se(x-y_).

        where t_{1-alpha/2} is the 100*(1-alpha/2) percentage point of
        the t-distribution with f degrees of freedom.

    References
    ----------
    Johnson, M G and Kotz, A, 1969, `The Encyclopedia of Statistics`
    (2), Griffin

    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    Snedecor, G W and Cochran, W G, 1967, `Statistical Methods`, Iowa
    State University Press
    """

    _return_scalars = _b_univar.ttest_2normal(
        tail,
        equal,
        nx,
        ny,
        xmean,
        ymean,
        xstd,
        ystd,
        clevel,
    )
    _t_py = _return_scalars.t
    _df_py = _return_scalars.df
    _prob_py = _return_scalars.prob
    _dl_py = _return_scalars.dl
    _du_py = _return_scalars.du
    _return_data = Ttest2normalReturnData(
        _t_py,
        _df_py,
        _prob_py,
        _dl_py,
        _du_py,
    )
    return _return_data

Robust1varMedianReturnData = _collections.namedtuple(
    'Robust1varMedianReturnData',
    ['y', 'xme', 'xmd', 'xsd'],
)

def robust_1var_median(x):
    u"""
    Robust estimation, median, median absolute deviation, robust
    standard deviation.

    ``robust_1var_median`` finds the median, median absolute deviation,
    and a robust estimate of the standard deviation for a set of
    ungrouped data.

    For full information please refer to the NAG Library document for
    g07da

    https://www.nag.com/numeric/nl/nagdoc_28.4/flhtml/g07/g07daf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The vector of observations, x_1,x_2,...,x_n.

    Returns
    -------
    y : float, ndarray, shape (n)
        The observations sorted into ascending order.

    xme : float
        The median, theta_med.

    xmd : float
        The median absolute deviation, sigma_med.

    xsd : float
        The robust estimate of the standard deviation, sigma'_med.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 1.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    The data consists of a sample of size n, denoted by x_1,x_2,...,x_n,
    drawn from a random variable X.

    ``robust_1var_median`` first computes the median,

        theta_med = med_i{x_i},

    and from this the median absolute deviation can be computed,

        sigma_med = med_i{|x_i-theta_med|}.

    Finally, a robust estimate of the standard deviation is computed,

        sigma'_med = sigma_med/Phi^-1(0.75)

    where Phi^-1(0.75) is the value of the inverse standard Normal
    function at the point 0.75.

    ``robust_1var_median`` is based upon function LTMDDV within the
    ROBETH library, see Marazzi (1987).

    References
    ----------
    Huber, P J, 1981, `Robust Statistics`, Wiley

    Marazzi, A, 1987, `Subroutines for robust estimation of location and
    scale in ROBETH`, Cah. Rech. Doc. IUMSP, No. 3 ROB 1, Institut
    Universitaire de Médecine Sociale et Préventive, Lausanne
    """

    fun_name = 'naginterfaces.library.univar.robust_1var_median'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_univar.robust_1var_median(
        _n,
        _x,
        _y,
    )
    _y_py = _y
    _xme_py = _return_scalars.xme
    _xmd_py = _return_scalars.xmd
    _xsd_py = _return_scalars.xsd
    _return_data = Robust1varMedianReturnData(
        _y_py,
        _xme_py,
        _xmd_py,
        _xsd_py,
    )
    return _return_data

Robust1varMestimReturnData = _collections.namedtuple(
    'Robust1varMestimReturnData',
    ['theta', 'sigma', 'rs', 'nit', 'wrk'],
)

def robust_1var_mestim(isigma, x, ipsi, c, h1, h2, h3, dchi, theta, sigma, tol, maxit=50):
    u"""
    Robust estimation, M-estimates for location and scale parameters,
    standard weight functions.

    ``robust_1var_mestim`` computes an M-estimate of location with
    (optional) simultaneous estimation of the scale using Huber's
    algorithm.

    For full information please refer to the NAG Library document for
    g07db

    https://www.nag.com/numeric/nl/nagdoc_28.4/flhtml/g07/g07dbf.html

    Parameters
    ----------
    isigma : int
        The value assigned to `isigma` determines whether sigma^ is to
        be simultaneously estimated.

        `isigma` = 0

            The estimation of sigma^ is bypassed and `sigma` is set
            equal to sigma_c.

        `isigma` = 1

            sigma^ is estimated simultaneously.

    x : float, array-like, shape (n)
        The vector of observations, x_1,x_2,...,x_n.

    ipsi : int
        Which psi function is to be used.

        `ipsi` = 0

            psi(t) = t.

        `ipsi` = 1

            Huber's function.

        `ipsi` = 2

            Hampel's piecewise linear function.

        `ipsi` = 3

            Andrew's sine wave,

        `ipsi` = 4

            Tukey's bi-weight.

    c : float
        If `ipsi` = 1, `c` must specify the parameter, c, of Huber's psi
        function. `c` is not referenced if `ipsi` != 1.

    h1 : float
        If `ipsi` = 2, `h1`, `h2` and `h3` must specify the parameters,
        h_1, h_2, and h_3, of Hampel's piecewise linear psi function.
        `h1`, `h2` and `h3` are not referenced if `ipsi` != 2.

    h2 : float
        If `ipsi` = 2, `h1`, `h2` and `h3` must specify the parameters,
        h_1, h_2, and h_3, of Hampel's piecewise linear psi function.
        `h1`, `h2` and `h3` are not referenced if `ipsi` != 2.

    h3 : float
        If `ipsi` = 2, `h1`, `h2` and `h3` must specify the parameters,
        h_1, h_2, and h_3, of Hampel's piecewise linear psi function.
        `h1`, `h2` and `h3` are not referenced if `ipsi` != 2.

    dchi : float
        d, the parameter of the chi function. `dchi` is not referenced
        if `ipsi` = 0.

    theta : float
        If `sigma` > 0 then `theta` must be set to the required starting
        value of the estimation of the location parameter theta^. A
        reasonable initial value for theta^ will often be the sample
        mean or median.

    sigma : float
        The role of `sigma` depends on the value assigned to `isigma`,
        as follows:

            if `isigma` = 1, `sigma` must be assigned a value which
            determines the values of the starting points for the
            calculations of theta^ and sigma^. If `sigma` <= 0.0 then
            ``robust_1var_mestim`` will determine the starting points of
            theta^ and sigma^. Otherwise the value assigned to `sigma`
            will be taken as the starting point for sigma^, and `theta`
            must be assigned a value before entry, see above;

            if `isigma` = 0, `sigma` must be assigned a value which
            determines the value of sigma_c, which is held fixed during
            the iterations, and the starting value for the calculation
            of theta^. If `sigma` <= 0, ``robust_1var_mestim`` will
            determine the value of sigma_c as the median absolute
            deviation adjusted to reduce bias (see
            ``robust_1var_median``) and the starting point for theta^.
            Otherwise, the value assigned to `sigma` will be taken as
            the value of sigma_c and `theta` must be assigned a relevant
            value before entry, see above.

    tol : float
        The relative precision for the final estimates. Convergence is
        assumed when the increments for `theta`, and `sigma` are less
        than `tol`*max(1.0, sigma_{k-1}).

    maxit : int, optional
        The maximum number of iterations that should be used during the
        estimation.

    Returns
    -------
    theta : float
        The M-estimate of the location parameter, theta^.

    sigma : float
        Contains the M-estimate of the scale parameter, sigma^, if
        `isigma` was assigned the value 1 on entry, otherwise `sigma`
        will contain the initial fixed value sigma_c.

    rs : float, ndarray, shape (n)
        The Winsorized residuals.

    nit : int
        The number of iterations that were used during the estimation.

    wrk : float, ndarray, shape (n)
        If `sigma` <= 0.0 on entry, `wrk` will contain the n
        observations in ascending order.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ipsi` = *<value>*.

            Constraint: `ipsi` = 0, 1, 2, 3 or 4.

        (`errno` 1)
            On entry, `isigma` = *<value>*.

            Constraint: `isigma` = 0 or 1.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 1.

        (`errno` 2)
            On entry, `dchi` = *<value>*.

            Constraint: `dchi` > 0.0.

        (`errno` 2)
            On entry, `h1` = *<value>*, `h2` = *<value>* and `h3` =
            *<value>*.

            Constraint: 0 <= `h1` <= `h2` <= `h3` and `h3` > 0.0.

        (`errno` 2)
            On entry, `c` = *<value>*.

            Constraint: `c` > 0.0.

        (`errno` 3)
            All elements of `x` are equal.

        (`errno` 4)
            Current estimate of `sigma` is zero or negative: `sigma` =
            *<value>*.

        (`errno` 5)
            Number of iterations required exceeds `maxit`: `maxit` =
            *<value>*.

        (`errno` 6)
            All winsorized residuals are zero.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    The data consists of a sample of size n, denoted by x_1,x_2,...,x_n,
    drawn from a random variable X.

    The x_i are assumed to be independent with an unknown distribution
    function of the form

        F((x_i-theta)/sigma)

    where theta is a location parameter, and sigma is a scale parameter.
    M-estimators of theta and sigma are given by the solution to the
    following system of equations:

        sum_i = 1^n psi((x_i-theta^)/sigma^) = 0

        sum_i = 1^n chi((x_i-theta^)/sigma^) = (n-1)beta

    where psi and chi are given functions, and beta is a constant, such
    that sigma^ is an unbiased estimator when x_i, for i = 1,2,...,n has
    a Normal distribution.
    Optionally, the second equation can be omitted and the first
    equation is solved for theta^ using an assigned value of sigma =
    sigma_c.

    The values of psi((x_i-theta^)/(sigma^))sigma^ are known as the
    Winsorized residuals.

    The following functions are available for psi and chi in
    ``robust_1var_mestim``:

    (a) **Null Weights**

        [table omitted]

        Use of these null functions leads to the mean and standard
        deviation of the data.

    (#) **Huber's Function**

        [table omitted]

    (#) **Hampel's Piecewise Linear Function**

        [table omitted]

    (#) **Andrew's Sine Wave Function**

        [table omitted]

    (#) **Tukey's Bi-weight**

        [table omitted]

    where c, h_1, h_2, h_3 and d are constants.

    Equations (1) and (2) are solved by a simple iterative procedure
    suggested by Huber:

        sigma^_k = sqrt(1/beta(n-1)(sum_i = 1^n
        chi((x_i-theta^_{k-1})/(sigma^_{k-1})))sigma^_k-1^2)

    and

        theta^_k = theta^_{k-1}+1/nsum_i = 1^n
        psi((x_i-theta^_{k-1})/(sigma^_k))sigma^_k

    or

        sigma^_k = sigma_c, if sigma is fixed.

    The initial values for theta^ and sigma^ may either be user-supplied
    or calculated within ``robust_1var_mestim`` as the sample median and
    an estimate of sigma based on the median absolute deviation
    respectively.

    ``robust_1var_mestim`` is based upon function LYHALG within the
    ROBETH library, see Marazzi (1987).

    References
    ----------
    Hampel, F R, Ronchetti, E M, Rousseeuw, P J and Stahel, W A, 1986,
    `Robust Statistics. The Approach Based on Influence Functions`,
    Wiley

    Huber, P J, 1981, `Robust Statistics`, Wiley

    Marazzi, A, 1987, `Subroutines for robust estimation of location and
    scale in ROBETH`, Cah. Rech. Doc. IUMSP, No. 3 ROB 1, Institut
    Universitaire de Médecine Sociale et Préventive, Lausanne
    """

    fun_name = 'naginterfaces.library.univar.robust_1var_mestim'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _rs = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _wrk = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_univar.robust_1var_mestim(
        isigma,
        _n,
        _x,
        ipsi,
        c,
        h1,
        h2,
        h3,
        dchi,
        theta,
        sigma,
        maxit,
        tol,
        _rs,
        _wrk,
    )
    _theta_py = _return_scalars.theta
    _sigma_py = _return_scalars.sigma
    _rs_py = _rs
    _nit_py = _return_scalars.nit
    _wrk_py = _wrk
    _return_data = Robust1varMestimReturnData(
        _theta_py,
        _sigma_py,
        _rs_py,
        _nit_py,
        _wrk_py,
    )
    return _return_data

Robust1varMestimWgtReturnData = _collections.namedtuple(
    'Robust1varMestimWgtReturnData',
    ['theta', 'sigma', 'rs', 'nit', 'wrk'],
)

def robust_1var_mestim_wgt(chi, psi, isigma, x, beta, theta, sigma, tol, maxit=50, data=None):
    u"""
    Robust estimation, M-estimates for location and scale parameters,
    user-defined weight functions.

    ``robust_1var_mestim_wgt`` computes an M-estimate of location with
    (optional) simultaneous estimation of scale, where you provide the
    weight functions.

    For full information please refer to the NAG Library document for
    g07dc

    https://www.nag.com/numeric/nl/nagdoc_28.4/flhtml/g07/g07dcf.html

    Parameters
    ----------
    chi : callable retval = chi(t, data=None)
        `chi` must return the value of the weight function chi for a
        given value of its argument.

        The value of chi must be non-negative.

        Parameters
        ~~~~~~~~~~
        t : float
            The argument for which `chi` must be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of the weight function phi evaluated at `t`.

    psi : callable retval = psi(t, data=None)
        `psi` must return the value of the weight function psi for a
        given value of its argument.

        Parameters
        ~~~~~~~~~~
        t : float
            The argument for which `psi` must be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of the weight function psi evaluated at `t`.

    isigma : int
        The value assigned to `isigma` determines whether sigma^ is to
        be simultaneously estimated.

        `isigma` = 0

            The estimation of sigma^ is bypassed and `sigma` is set
            equal to sigma_c.

        `isigma` = 1

            sigma^ is estimated simultaneously.

    x : float, array-like, shape (n)
        The vector of observations, x_1,x_2,...,x_n.

    beta : float
        The value of the constant beta of the chosen `chi` function.

    theta : float
        If `sigma` > 0, `theta` must be set to the required starting
        value of the estimate of the location parameter theta^. A
        reasonable initial value for theta^ will often be the sample
        mean or median.

    sigma : float
        The role of `sigma` depends on the value assigned to `isigma` as
        follows.

        If `isigma` = 1, `sigma` must be assigned a value which
        determines the values of the starting points for the calculation
        of theta^ and sigma^.

        If `sigma` <= 0.0, ``robust_1var_mestim_wgt`` will determine the
        starting points of theta^ and sigma^.

        Otherwise, the value assigned to `sigma` will be taken as the
        starting point for sigma^, and `theta` must be assigned a
        relevant value before entry, see above.

        If `isigma` = 0, `sigma` must be assigned a value which
        determines the values of sigma_c, which is held fixed during the
        iterations, and the starting value for the calculation of
        theta^.

        If `sigma` <= 0, ``robust_1var_mestim_wgt`` will determine the
        value of sigma_c as the median absolute deviation adjusted to
        reduce bias (see ``robust_1var_median``) and the starting point
        for theta.

        Otherwise, the value assigned to `sigma` will be taken as the
        value of sigma_c and `theta` must be assigned a relevant value
        before entry, see above.

    tol : float
        The relative precision for the final estimates. Convergence is
        assumed when the increments for `theta`, and `sigma` are less
        than `tol`*max(1.0, sigma_{k-1}).

    maxit : int, optional
        The maximum number of iterations that should be used during the
        estimation.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    theta : float
        The M-estimate of the location parameter theta^.

    sigma : float
        The M-estimate of the scale parameter sigma^, if `isigma` was
        assigned the value 1 on entry, otherwise `sigma` will contain
        the initial fixed value sigma_c.

    rs : float, ndarray, shape (n)
        The Winsorized residuals.

    nit : int
        The number of iterations that were used during the estimation.

    wrk : float, ndarray, shape (n)
        If `sigma` <= 0.0 on entry, `wrk` will contain the n
        observations in ascending order.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `isigma` = *<value>*.

            Constraint: `isigma` = 0 or 1.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 1.

        (`errno` 2)
            On entry, `beta` = *<value>*.

            Constraint: `beta` > 0.0.

        (`errno` 3)
            All elements of `x` are equal.

        (`errno` 4)
            Current estimate of `sigma` is zero or negative: `sigma` =
            *<value>*.

        (`errno` 5)
            Number of iterations required exceeds `maxit`: `maxit` =
            *<value>*.

        (`errno` 6)
            All winsorized residuals are zero.

        (`errno` 7)
            The `chi` function returned a negative value: `chi` =
            *<value>*.

    Notes
    -----
    The data consists of a sample of size n, denoted by x_1,x_2,...,x_n,
    drawn from a random variable X.

    The x_i are assumed to be independent with an unknown distribution
    function of the form,

        F((x_i-theta)/sigma)

    where theta is a location parameter, and sigma is a scale parameter.
    M-estimators of theta and sigma are given by the solution to the
    following system of equations;

        [table omitted]

    where psi and chi are user-supplied weight functions, and beta is a
    constant.
    Optionally the second equation can be omitted and the first equation
    is solved for theta^ using an assigned value of sigma = sigma_c.

    The constant beta should be chosen so that sigma^ is an unbiased
    estimator when x_i, for i = 1,2,...,n has a Normal distribution.
    To achieve this the value of beta is calculated as:

        beta = E(chi) = int_- infinity^infinity chi(z)1/(sqrt(2
        pi))exp{(-z^2)/2}dz

    The values of psi((x_i-theta^)/(sigma^))sigma^ are known as the
    Winsorized residuals.

    The equations are solved by a simple iterative procedure, suggested
    by Huber:

        sigma^_k = sqrt(1/beta(n-1)(sum_i = 1^n
        chi((x_i-theta^_{k-1})/(sigma^_{k-1})))sigma^_k-1^2)

    and

        theta^_k = theta^_{k-1}+1/nsum_i = 1^n
        psi((x_i-theta^_{k-1})/(sigma^_k))sigma^_k

    or

        sigma^_k = sigma_c

    if sigma is fixed.

    The initial values for theta^ and sigma^ may be user-supplied or
    calculated within ``robust_1var_mestim`` as the sample median and an
    estimate of sigma based on the median absolute deviation
    respectively.

    ``robust_1var_mestim_wgt`` is based upon function LYHALG within the
    ROBETH library, see Marazzi (1987).

    References
    ----------
    Hampel, F R, Ronchetti, E M, Rousseeuw, P J and Stahel, W A, 1986,
    `Robust Statistics. The Approach Based on Influence Functions`,
    Wiley

    Huber, P J, 1981, `Robust Statistics`, Wiley

    Marazzi, A, 1987, `Subroutines for robust estimation of location and
    scale in ROBETH`, Cah. Rech. Doc. IUMSP, No. 3 ROB 1, Institut
    Universitaire de Médecine Sociale et Préventive, Lausanne
    """

    def cb_chi(t, data=None):
        """Wrapping user's chi. Called as base chi."""
        if data is not None:
            _retval_chi = chi(
                t,
                data,
            )
        else:
            _retval_chi = chi(
                t,
            )
        return _retval_chi

    def cb_psi(t, data=None):
        """Wrapping user's psi. Called as base psi."""
        if data is not None:
            _retval_psi = psi(
                t,
                data,
            )
        else:
            _retval_psi = psi(
                t,
            )
        return _retval_psi

    fun_name = 'naginterfaces.library.univar.robust_1var_mestim_wgt'

    _b_utils._check_callable(chi, {'fun_name': fun_name, 'entity': 'chi'})
    _chi = cb_chi
    _b_utils._check_callable(psi, {'fun_name': fun_name, 'entity': 'psi'})
    _psi = cb_psi
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _rs = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _wrk = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_univar.robust_1var_mestim_wgt(
        _chi,
        _psi,
        isigma,
        _n,
        _x,
        beta,
        theta,
        sigma,
        maxit,
        tol,
        _rs,
        _wrk,
        data,
    )
    _theta_py = _return_scalars.theta
    _sigma_py = _return_scalars.sigma
    _rs_py = _rs
    _nit_py = _return_scalars.nit
    _wrk_py = _wrk
    _return_data = Robust1varMestimWgtReturnData(
        _theta_py,
        _sigma_py,
        _rs_py,
        _nit_py,
        _wrk_py,
    )
    return _return_data

Robust1varTrimmedReturnData = _collections.namedtuple(
    'Robust1varTrimmedReturnData',
    ['tmean', 'wmean', 'tvar', 'wvar', 'k', 'sx'],
)

def robust_1var_trimmed(x, alpha):
    u"""
    Computes a trimmed and winsorized mean of a single sample with
    estimates of their variance.

    ``robust_1var_trimmed`` calculates the trimmed and Winsorized means
    of a sample and estimates of the variances of the two means.

    For full information please refer to the NAG Library document for
    g07dd

    https://www.nag.com/numeric/nl/nagdoc_28.4/flhtml/g07/g07ddf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The sample observations, x_i, for i = 1,2,...,n.

    alpha : float
        alpha, the proportion of observations to be trimmed at each end
        of the sorted sample.

    Returns
    -------
    tmean : float
        The alpha-trimmed mean, x__t.

    wmean : float
        The alpha-Winsorized mean, x__w.

    tvar : float
        Contains an estimate of the variance of the trimmed mean.

    wvar : float
        Contains an estimate of the variance of the Winsorized mean.

    k : int
        Contains the number of observations trimmed at each end, k.

    sx : float, ndarray, shape (n)
        Contains the sample observations sorted into ascending order.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n <= 1.

        (`errno` 2)
            On entry, `alpha` = *<value>*.

            Constraint: 0.0 <= `alpha` < 0.5.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``robust_1var_trimmed`` calculates the alpha-trimmed mean and
    alpha-Winsorized mean for a given alpha, as described below.

    Let x_i, for i = 1,2,...,n represent the n sample observations
    sorted into ascending order.
    Let k = [alpha n] where [y] represents the integer nearest to y; if
    2k = n then k is reduced by 1.

    Then the trimmed mean is defined as:

        x__t = 1/(n-2k)sum_i = k+1^n-kx_i,

    and the Winsorized mean is defined as:

        x__w = 1/n(sum_i = k+1^n-kx_i+k*x_{k+1}+k*x_{n-k}).

    ``robust_1var_trimmed`` then calculates the Winsorized variance
    about the trimmed and Winsorized means respectively and divides by n
    to obtain estimates of the variances of the above two means.

    Thus we have;

        Estimate of var(x__t) = 1/(n^2)(sum_i =
        k+1^n-k(x_i-x__t)^2+k(x_{k+1}-x__t)^2+k(x_{n-k}-x__t)^2)

    and

        Estimate of var(x__w) = 1/(n^2)(sum_i =
        k+1^n-k(x_i-x__w)^2+k(x_{k+1}-x__w)^2+k(x_{n-k}-x__w)^2).

    References
    ----------
    Hampel, F R, Ronchetti, E M, Rousseeuw, P J and Stahel, W A, 1986,
    `Robust Statistics. The Approach Based on Influence Functions`,
    Wiley

    Huber, P J, 1981, `Robust Statistics`, Wiley
    """

    fun_name = 'naginterfaces.library.univar.robust_1var_trimmed'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _sx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_univar.robust_1var_trimmed(
        _n,
        _x,
        alpha,
        _sx,
    )
    _tmean_py = _return_scalars.tmean
    _wmean_py = _return_scalars.wmean
    _tvar_py = _return_scalars.tvar
    _wvar_py = _return_scalars.wvar
    _k_py = _return_scalars.k
    _sx_py = _sx
    _return_data = Robust1varTrimmedReturnData(
        _tmean_py,
        _wmean_py,
        _tvar_py,
        _wvar_py,
        _k_py,
        _sx_py,
    )
    return _return_data

Robust1varCiReturnData = _collections.namedtuple(
    'Robust1varCiReturnData',
    ['theta', 'thetal', 'thetau', 'estcl', 'wlower', 'wupper'],
)

def robust_1var_ci(method, x, clevel):
    u"""
    Robust confidence intervals, one-sample.

    ``robust_1var_ci`` computes a rank based (nonparametric) estimate
    and confidence interval for the location parameter of a single
    population.

    For full information please refer to the NAG Library document for
    g07ea

    https://www.nag.com/numeric/nl/nagdoc_28.4/flhtml/g07/g07eaf.html

    Parameters
    ----------
    method : str, length 1
        Specifies the method to be used.

        `method` = 'E'

            The exact algorithm is used.

        `method` = 'A'

            The iterative algorithm is used.

    x : float, array-like, shape (n)
        The sample observations, x_i, for i = 1,2,...,n.

    clevel : float
        The confidence interval desired.

        For example, for a 95% confidence interval set `clevel` = 0.95.

    Returns
    -------
    theta : float
        The estimate of the location, theta^.

    thetal : float
        The estimate of the lower limit of the confidence interval,
        theta_l.

    thetau : float
        The estimate of the upper limit of the confidence interval,
        theta_u.

    estcl : float
        An estimate of the actual percentage confidence of the interval
        found, as a proportion between (0.0,1.0).

    wlower : float
        The upper value of the Wilcoxon test statistic, W_u,
        corresponding to the lower limit of the confidence interval.

    wupper : float
        The lower value of the Wilcoxon test statistic, W_l,
        corresponding to the upper limit of the confidence interval.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `clevel` = *<value>*.

            Constraint: 0.0 < `clevel` < 1.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'E' or 'A'.

        (`errno` 2)
            Not enough information to compute an interval estimate since
            the whole sample is identical. The common value is returned
            in `theta`, `thetal` and `thetau`.

        (`errno` 3)
            The iterative procedure used to estimate theta has not
            converged.

        (`errno` 3)
            The iterative procedure used to estimate, theta_u, the upper
            confidence limit has not converged.

        (`errno` 3)
            The iterative procedure used to estimate, theta_l, the lower
            confidence limit has not converged.

    Notes
    -----
    Consider a vector of independent observations, x =
    (x_1,x_2,...,x_n)^T with unknown common symmetric density
    f(x_i-theta). ``robust_1var_ci`` computes the Hodges--Lehmann
    location estimator (see Lehmann (1975)) of the centre of symmetry
    theta, together with an associated confidence interval.
    The Hodges--Lehmann estimate is defined as

        theta^= median{(x_i+x_j)/2,1 <= i <= j <= n}.

    Let m = (n(n+1))/2 and let a_k, for k = 1,2,...,m denote the m
    ordered averages (x_i+x_j)/2 for 1 <= i <= j <= n.
    Then

        if m is odd, theta^= a_k where k = (m+1)/2;

        if m is even, theta^= (a_k+a_{k+1})/2 where k = m/2.

    This estimator arises from inverting the one-sample Wilcoxon
    signed-rank test statistic, W(x-theta_0), for testing the hypothesis
    that theta = theta_0.
    Effectively W(x-theta_0) is a monotonically decreasing step function
    of theta_0 with

        mean (W) = mu = n(n+1)/4,;;;var(W) = sigma^2 = n(n+1)(2n+1)/24.

    The estimate theta^ is the solution to the equation W(x-theta^) =
    mu; two methods are available for solving this equation.
    These methods avoid the computation of all the ordered averages a_k;
    this is because for large n both the storage requirements and the
    computation time would be excessive.

    The first is an exact method based on a set partitioning procedure
    on the set of all ordered averages (x_i+x_j)/2 for i <= j.
    This is based on the algorithm proposed by Monahan (1984).

    The second is an iterative algorithm, based on the Illinois method
    which is a modification of the `regula falsi` method, see McKean and
    Ryan (1977).
    This algorithm has proved suitable for the function W(x-theta_0)
    which is asymptotically linear as a function of theta_0.

    The confidence interval limits are also based on the inversion of
    the Wilcoxon test statistic.

    Given a desired percentage for the confidence interval, 1-alpha,
    expressed as a proportion between 0 and 1, initial estimates for the
    lower and upper confidence limits of the Wilcoxon statistic are
    found from

        W_l = mu-0.5+(sigmaPhi^-1(alpha/2))

    and

        W_u = mu+0.5+(sigmaPhi^-1(1-alpha/2)),

    where Phi^-1 is the inverse cumulative Normal distribution function.

    W_l and W_u are rounded to the nearest integer values.
    These estimates are then refined using an exact method if n <= 80,
    and a Normal approximation otherwise, to find W_l and W_u satisfying

        P(W <= W_l) <= alpha/2;P(W <= W_l+1) > alpha/2

    and

        P(W >= W_u) <= alpha /2;P(W >= W_u-1) > alpha /2.

    Let W_u = m-k; then theta_l = a_{k+1}.
    This is the largest value theta_l such that W(x-theta_l) = W_u.

    Let W_l = k; then theta_u = a_{m-k}.
    This is the smallest value theta_u such that W(x-theta_u) = W_l.

    As in the case of theta^, these equations may be solved using either
    the exact or the iterative methods to find the values theta_l and
    theta_u.

    Then (theta_l,theta_u) is the confidence interval for theta.
    The confidence interval is thus defined by those values of theta_0
    such that the null hypothesis, theta = theta_0, is not rejected by
    the Wilcoxon signed-rank test at the (100*alpha)% level.

    References
    ----------
    Lehmann, E L, 1975, `Nonparametrics: Statistical Methods Based on
    Ranks`, Holden--Day

    Marazzi, A, 1987, `Subroutines for robust estimation of location and
    scale in ROBETH`, Cah. Rech. Doc. IUMSP, No. 3 ROB 1, Institut
    Universitaire de Médecine Sociale et Préventive, Lausanne

    McKean, J W and Ryan, T A, 1977, `Algorithm 516: An algorithm for
    obtaining confidence intervals and point estimates based on ranks in
    the two-sample location problem`, ACM Trans. Math. Software (10),
    183--185

    Monahan, J F, 1984, `Algorithm 616: Fast computation of the
    Hodges--Lehman location estimator`, ACM Trans. Math. Software (10),
    265--270
    """

    fun_name = 'naginterfaces.library.univar.robust_1var_ci'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalars = _b_univar.robust_1var_ci(
        method,
        _n,
        _x,
        clevel,
    )
    _theta_py = _return_scalars.theta
    _thetal_py = _return_scalars.thetal
    _thetau_py = _return_scalars.thetau
    _estcl_py = _return_scalars.estcl
    _wlower_py = _return_scalars.wlower
    _wupper_py = _return_scalars.wupper
    _return_data = Robust1varCiReturnData(
        _theta_py,
        _thetal_py,
        _thetau_py,
        _estcl_py,
        _wlower_py,
        _wupper_py,
    )
    return _return_data

Robust2varCiReturnData = _collections.namedtuple(
    'Robust2varCiReturnData',
    ['theta', 'thetal', 'thetau', 'estcl', 'ulower', 'uupper'],
)

def robust_2var_ci(method, x, y, clevel):
    u"""
    Robust confidence intervals, two-sample.

    ``robust_2var_ci`` calculates a rank based (nonparametric) estimate
    and confidence interval for the difference in location between two
    independent populations.

    For full information please refer to the NAG Library document for
    g07eb

    https://www.nag.com/numeric/nl/nagdoc_28.4/flhtml/g07/g07ebf.html

    Parameters
    ----------
    method : str, length 1
        Specifies the method to be used.

        `method` = 'E'

            The exact algorithm is used.

        `method` = 'A'

            The iterative algorithm is used.

    x : float, array-like, shape (n)
        The observations of the first sample, x_i, for i = 1,2,...,n.

    y : float, array-like, shape (m)
        The observations of the second sample, y_j, for j = 1,2,...,m.

    clevel : float
        The confidence interval required, 1-alpha; e.g., for a 95%
        confidence interval set `clevel` = 0.95.

    Returns
    -------
    theta : float
        The estimate of the difference in the location of the two
        populations, theta^.

    thetal : float
        The estimate of the lower limit of the confidence interval,
        theta_l.

    thetau : float
        The estimate of the upper limit of the confidence interval,
        theta_u.

    estcl : float
        An estimate of the actual percentage confidence of the interval
        found, as a proportion between (0.0,1.0).

    ulower : float
        The value of the Mann--Whitney U statistic corresponding to the
        lower confidence limit, U_l.

    uupper : float
        The value of the Mann--Whitney U statistic corresponding to the
        upper confidence limit, U_u.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 1)
            On entry, `clevel` = *<value>*.

            Constraint: 0.0 < `clevel` < 1.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'E' or 'A'.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Not enough information to compute an interval estimate since
            each sample has identical values. The common difference is
            returned in `theta`, `thetal` and `thetau`.

        (`errno` 3)
            The iterative procedure used to estimate theta has not
            converged.

        (`errno` 3)
            The iterative procedure used to estimate, theta_u, the upper
            confidence limit has not converged.

        (`errno` 3)
            The iterative procedure used to estimate, theta_l, the lower
            confidence limit has not converged.

    Notes
    -----
    Consider two random samples from two populations which have the same
    continuous distribution except for a shift in the location.
    Let the random sample, x = (x_1,x_2,...,x_n)^T, have distribution
    F(x) and the random sample, y = (y_1,y_2,...,y_m)^T, have
    distribution F(x-theta).

    ``robust_2var_ci`` finds a point estimate, theta^, of the difference
    in location theta together with an associated confidence interval.
    The estimates are based on the ordered differences y_j-x_i.
    The estimate theta^ is defined by

        theta^= median{y_j-x_i, i = 1,2,...,n;j = 1,2,...,m}.

    Let d_k, for k = 1,2,...,nm, denote the nm (ascendingly) ordered
    differences y_j-x_i, for j = 1,2,...,m, for i = 1,2,...,n.
    Then

        if nm is odd, theta^= d_k where k = (nm-1)/2;

        if nm is even, theta^= (d_k+d_{k+1})/2 where k = nm/2.

    This estimator arises from inverting the two sample Mann--Whitney
    rank test statistic, U(theta_0), for testing the hypothesis that
    theta = theta_0.
    Thus U(theta_0) is the value of the Mann--Whitney U statistic for
    the two independent samples {(x_i+theta_0), for i = 1,2,...,n} and
    {y_j, for j = 1,2,...,m}.
    Effectively U(theta_0) is a monotonically increasing step function
    of theta_0 with

        mean (U) = mu = nm/2,;;var(U) = sigma^2nm(n+m+1)/12.

    The estimate theta^ is the solution to the equation U(theta^) = mu;
    two methods are available for solving this equation.
    These methods avoid the computation of all the ordered differences
    d_k; this is because for large n and m both the storage requirements
    and the computation time would be high.

    The first is an exact method based on a set partitioning procedure
    on the set of all differences y_j-x_i, for j = 1,2,...,m, for i =
    1,2,...,n.
    This is adapted from the algorithm proposed by Monahan (1984) for
    the computation of the Hodges--Lehmann estimator for a single
    population.

    The second is an iterative algorithm, based on the Illinois method
    which is a modification of the `regula falsi` method, see McKean and
    Ryan (1977).
    This algorithm has proved suitable for the function U(theta_0) which
    is asymptotically linear as a function of theta_0.

    The confidence interval limits are also based on the inversion of
    the Mann--Whitney test statistic.

    Given a desired percentage for the confidence interval, 1-alpha,
    expressed as a proportion between 0.0 and 1.0 initial estimates of
    the upper and lower confidence limits for the Mann--Whitney U
    statistic are found;

        U_l = mu-0.5+(sigma*Phi^-1(alpha/2));;U_u =
        mu+0.5+(sigma*Phi^-1((1-alpha)/2))

    where Phi^-1 is the inverse cumulative Normal distribution function.

    U_l and U_u are rounded to the nearest integer values.
    These estimates are refined using an exact method, without taking
    ties into account, if n+m <= 40 and max(n, m) <= 30 and a Normal
    approximation otherwise, to find U_l and U_u satisfying

        P(U <= U_l) <= alpha/2;P(U <= U_l+1) > alpha/2

    and

        P(U >= U_u) <= alpha /2;P(U >= U_u-1) > alpha /2.

    The function U(theta_0) is a monotonically increasing step function.
    It is the number of times a score in the second sample, y_j,
    precedes a score in the first sample, x_i+theta, where we only count
    a half if a score in the second sample actually equals a score in
    the first.

    Let U_l = k; then theta_l = d_{k+1}.
    This is the largest value theta_l such that U(theta_l) = U_l.

    Let U_u = nm-k; then theta_u = d_{nm-k}.
    This is the smallest value theta_u such that U(theta_u) = U_u.

    As in the case of theta^, these equations may be solved using either
    the exact or iterative methods to find the values theta_l and
    theta_u.

    Then (theta_l,theta_u) is the confidence interval for theta.
    The confidence interval is thus defined by those values of theta_0
    such that the null hypothesis, theta = theta_0, is not rejected by
    the Mann--Whitney two sample rank test at the (100*alpha)% level.

    References
    ----------
    Lehmann, E L, 1975, `Nonparametrics: Statistical Methods Based on
    Ranks`, Holden--Day

    McKean, J W and Ryan, T A, 1977, `Algorithm 516: An algorithm for
    obtaining confidence intervals and point estimates based on ranks in
    the two-sample location problem`, ACM Trans. Math. Software (10),
    183--185

    Monahan, J F, 1984, `Algorithm 616: Fast computation of the
    Hodges--Lehman location estimator`, ACM Trans. Math. Software (10),
    265--270
    """

    fun_name = 'naginterfaces.library.univar.robust_2var_ci'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    try:
        _return_scalars = _b_univar.robust_2var_ci(
            method,
            _n,
            _x,
            _m,
            _y,
            clevel,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _theta_py = _return_scalars.theta
    _thetal_py = _return_scalars.thetal
    _thetau_py = _return_scalars.thetau
    _estcl_py = _return_scalars.estcl
    _ulower_py = _return_scalars.ulower
    _uupper_py = _return_scalars.uupper
    _return_data = Robust2varCiReturnData(
        _theta_py,
        _thetal_py,
        _thetau_py,
        _estcl_py,
        _ulower_py,
        _uupper_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

OutlierPeirce1varReturnData = _collections.namedtuple(
    'OutlierPeirce1varReturnData',
    ['iout', 'niout', 'diff', 'llamb'],
)

def outlier_peirce_1var(p, y, ldiff, mean=0.0, var=0.0):
    u"""
    Outlier detection using method of Peirce, raw data or single
    variance supplied.

    ``outlier_peirce_1var`` identifies outlying values using Peirce's
    criterion.

    For full information please refer to the NAG Library document for
    g07ga

    https://www.nag.com/numeric/nl/nagdoc_28.4/flhtml/g07/g07gaf.html

    Parameters
    ----------
    p : int
        p, the number of parameters in the model used in obtaining the
        y. If y is an observed set of values, as opposed to the
        residuals from fitting a model with p parameters, then p should
        be set to 1, i.e., as if a model just containing the mean had
        been used.

    y : float, array-like, shape (n)
        y, the data being tested.

    ldiff : int
        The maximum number of values to be returned in arrays `diff` and
        `llamb`.

        If `ldiff` <= 0, arrays `diff` and `llamb` are not referenced.

    mean : float, optional
        If `var` > 0.0, `mean` must contain mu, the mean of y, otherwise
        `mean` is not referenced and the mean is calculated from the
        data supplied in `y`.

    var : float, optional
        If `var` > 0.0, `var` must contain sigma^2, the variance of y,
        otherwise the variance is calculated from the data supplied in
        `y`.

    Returns
    -------
    iout : int, ndarray, shape (n)
        The indices of the values in `y` sorted in descending order of
        the absolute difference from the mean, therefore, |`y`[
        `iout`[i-2] -1]-mu| >= |`y`[ `iout`[i-1] -1]-mu|, for i =
        2,3,...,n.

    niout : int
        The number of potential outliers. The indices for these
        potential outliers are held in the first `niout` elements of
        `iout`. By construction there can be at most n-`p`-1 values
        flagged as outliers.

    diff : float, ndarray, shape (`ldiff`)
        `diff`[i-1] holds |y-mu|-sigma^2z for observation
        `y`[`iout`[i-1]-1], for i = 1,2,...,min(`ldiff`, `niout`+1,
        n-`p`-1).

    llamb : float, ndarray, shape (`ldiff`)
        `llamb`[i-1] holds log(lambda^2) for observation
        `y`[`iout`[i-1]-1], for i = 1,2,...,min(`ldiff`, `niout`+1,
        n-`p`-1).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 3.

        (`errno` 2)
            On entry, `p` = *<value>* and n = *<value>*.

            Constraint: 1 <= `p` <= n-2.

    Notes
    -----
    ``outlier_peirce_1var`` flags outlying values in data using Peirce's
    criterion.
    Let

        y denote a vector of n observations (for example the residuals)
        obtained from a model with p parameters,

        m denote the number of potential outlying values,

        mu and sigma^2 denote the mean and variance of y respectively,

        y~ denote a vector of length n-m constructed by dropping the m
        values from y with the largest value of |y_i-mu|,

        sigma~^2 denote the (unknown) variance of y~,

        lambda denote the ratio of sigma~ and sigma with lambda =
        (sigma~)/sigma.

    Peirce's method flags y_i as a potential outlier if |y_i-mu| >= x,
    where x = sigma^2z and z is obtained from the solution of

        R^m = lambda^(m-n)m^m(n-m)^(n-m)/(n^n)

    where

        R = 2exp(((z^2-1)/2)(1-Phi(z)))

    and Phi is the cumulative distribution function for the standard
    Normal distribution.

    As sigma~^2 is unknown an assumption is made that the relationship
    between sigma~^2 and sigma^2, hence lambda, depends only on the sum
    of squares of the rejected observations and the ratio estimated as

        lambda^2 = (n-p-mz^2)/(n-p-m)

    which gives

        z^2 = 1+(n-p-m)/m(1-lambda^2)

    A value for the cutoff x is calculated iteratively.
    An initial value of R = 0.2 is used and a value of lambda is
    estimated using equation [equation].
    Equation [equation] is then used to obtain an estimate of z and then
    equation [equation] is used to get a new estimate for R.
    This process is then repeated until the relative change in z between
    consecutive iterations is <= sqrt(epsilon), where epsilon is machine
    precision.

    By construction, the cutoff for testing for m+1 potential outliers
    is less than the cutoff for testing for m potential outliers.
    Therefore, Peirce's criterion is used in sequence with the existence
    of a single potential outlier being investigated first.
    If one is found, the existence of two potential outliers is
    investigated etc.

    If one of a duplicate series of observations is flagged as an
    outlier, then all of them are flagged as outliers.

    References
    ----------
    Gould, B A, 1855, `On Peirce's criterion for the rejection of
    doubtful observations, with tables for facilitating its
    application`, The Astronomical Journal (45)

    Peirce, B, 1852, `Criterion for the rejection of doubtful
    observations`, The Astronomical Journal (45)
    """

    fun_name = 'naginterfaces.library.univar.outlier_peirce_1var'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _iout = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    _ldiff = _b_utils._EngineIntScalarType(
        ldiff, {'fun_name': fun_name, 'entity': 'ldiff'},
    ).to_ctypes()
    _ldiff_py = _b_utils._EngineIntScalarType.to_py(_ldiff)
    _diff = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ldiff_py,),
    )
    _llamb = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ldiff_py,),
    )
    _return_scalar = _b_univar.outlier_peirce_1var(
        _n,
        p,
        _y,
        mean,
        var,
        _iout,
        _ldiff,
        _diff,
        _llamb,
    )
    _iout_py = _iout
    _niout_py = _return_scalar
    _diff_py = _diff
    _llamb_py = _llamb
    _return_data = OutlierPeirce1varReturnData(
        _iout_py,
        _niout_py,
        _diff_py,
        _llamb_py,
    )
    return _return_data

OutlierPeirce2varReturnData = _collections.namedtuple(
    'OutlierPeirce2varReturnData',
    ['outl', 'x', 'lx', 'ux'],
)

def outlier_peirce_2var(n, e, var1, var2):
    u"""
    Outlier detection using method of Peirce, two variances supplied.

    ``outlier_peirce_2var`` returns a flag indicating whether a single
    data point is an outlier as defined by Peirce's criterion.

    For full information please refer to the NAG Library document for
    g07gb

    https://www.nag.com/numeric/nl/nagdoc_28.4/flhtml/g07/g07gbf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    e : float
        e~, the value being tested.

    var1 : float
        sigma^2, the residual variance on fitting model M to y.

    var2 : float
        sigma~^2, the residual variance on fitting model M to y~.

    Returns
    -------
    outl : bool
        True if the value being tested is an outlier.

    x : float
        An estimated value of x, the cutoff that indicates an outlier.

    lx : float
        l, the lower limit for x.

    ux : float
        u, the upper limit for x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 3.

        (`errno` 3)
            On entry, `var1` = *<value>*.

            Constraint: `var1` > 0.0.

        (`errno` 4)
            On entry, `var2` = *<value>*.

            Constraint: `var2` > 0.0.

        (`errno` 4)
            On entry, `var1` = *<value>*, `var2` = *<value>*.

            Constraint: `var2` < `var1`.

    Notes
    -----
    ``outlier_peirce_2var`` tests a potential outlying value using
    Peirce's criterion.
    Let

        e denote a vector of n residuals with mean zero and variance
        sigma^2 obtained from fitting some model M to a series of data
        y,

        e~ denote the largest absolute residual in e, i.e., |e~| >=
        |e_i| for all i, and let y~ denote the data series y with the
        observation corresponding to e~ having been omitted,

        sigma~^2 denote the residual variance on fitting model M to y~,

        lambda denote the ratio of sigma~ and sigma with lambda =
        (sigma~)/sigma.

    Peirce's method flags e~ as a potential outlier if |e~| >= x, where
    x = sigma^2z and z is obtained from the solution of

        R = lambda^(1-n)(n-1)^(n-1)/(n^n)

    where

        R = 2exp(((z^2-1)/2)(1-Phi(z)))

    and Phi is the cumulative distribution function for the standard
    Normal distribution.

    Unlike ``outlier_peirce_1var``, both sigma^2 and sigma~^2 must be
    supplied and, therefore, no assumptions are made about the nature of
    the relationship between these two quantities.
    Only a single potential outlier is tested for at a time.

    This function uses an algorithm described in ``opt.one_var_func`` to
    refine a lower, l, and upper, u, limit for x.
    This refinement stops when |e~| < l or |e~| > u.

    References
    ----------
    Gould, B A, 1855, `On Peirce's criterion for the rejection of
    doubtful observations, with tables for facilitating its
    application`, The Astronomical Journal (45)

    Peirce, B, 1852, `Criterion for the rejection of doubtful
    observations`, The Astronomical Journal (45)
    """

    _return_scalars = _b_univar.outlier_peirce_2var(
        n,
        e,
        var1,
        var2,
    )
    _outl_py = _return_scalars.outl
    _x_py = _return_scalars.x
    _lx_py = _return_scalars.lx
    _ux_py = _return_scalars.ux
    _return_data = OutlierPeirce2varReturnData(
        _outl_py,
        _x_py,
        _lx_py,
        _ux_py,
    )
    return _return_data
